#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# utils.py
# ffsas: free-form inversion for small-angle scattering
# Copyright © 2021 SciML, STFC, UK. All rights reserved.


""" utilities """

import logging
import sys
import time
import uuid
from contextlib import contextmanager
from pathlib import Path

import numpy as np


class MultiLevelLogger:
    """ Class to log multiple levels of sub-processes """

    class SimpleTimer:
        """ Simple timer class """

        def __init__(self, proc_name):
            self.start_time = time.time()
            self.proc_name = proc_name

        def elapsed(self):
            return time.time() - self.start_time

    def __init__(self, indent_char='.', indent_width=4):
        """
        Create a multi-level logger

        :param indent_char: character for indent fill (default: '.')
        :param indent_width: indent width (default: 4)
        """
        # indent
        self._indent_char = indent_char
        self._indent_width = indent_width

        # to be initialized later in activate() because
        # log activation is rank-dependent
        self._timers = []
        self._logger = None

    def activate(self, name=None, file_path=None, screen=True):
        """
        Activate this logger

        :param name: name of the logger (default: None)
        :param file_path: file to save logs (default: None)
        :param screen: show logs on screen (default: True)
        """
        # random name
        if name is None:
            name = str(uuid.uuid4())
        self._logger = logging.getLogger(name)
        self._logger.setLevel(logging.INFO)
        if file_path is not None:
            self._logger.addHandler(
                logging.FileHandler(Path(file_path).expanduser(), 'w'))
        if screen:
            self._logger.addHandler(logging.StreamHandler(sys.stdout))

    @property
    def activated(self):
        return self._logger is not None

    @property
    def current_level(self):
        return len(self._timers)

    @property
    def elapsed_shallowest(self):
        """ Elapsed time of the shallowest level """
        if self.current_level == 0:
            return 0.  # not activated or not called start()
        else:
            return self._timers[-1].elapsed()

    def begin(self, proc_name: str):
        """
        Begin a sub-process

        :param proc_name: name of the sub-process
        """
        if not self.activated:
            return

        # message
        message = [self._indent_char * self._indent_width * self.current_level,
                   '<BEGIN> ', proc_name]
        self._logger.info(''.join(message))
        # timer
        self._timers.append(MultiLevelLogger.SimpleTimer(proc_name))

    def ended(self, proc_name: str = ''):
        """
        End a sub-process

        :param proc_name: name of the sub-process
        """
        if not self.activated:
            return

        # timer
        timer = self._timers.pop()
        elapsed = timer.elapsed()
        # message
        if proc_name == '':
            # calling ended() without proc_name
            proc_name = timer.proc_name
        else:
            # calling ended() with proc_name, check consistency
            assert proc_name == timer.proc_name, \
                f"Subprocess names do not match in " \
                f"begin() and ended() of a MultiLevelLogger instance." \
                f"\nSubprocess name passed to begin(): {timer.proc_name}" \
                f"\nSubprocess name passed to ended(): {proc_name}"
        message = [self._indent_char * self._indent_width * self.current_level,
                   '<ENDED> ', proc_name, f' [ELAPSED = {elapsed:f} sec]']
        self._logger.info(''.join(message))

    def message(self, what: str):
        """
        Log a message to the current sub-process

        :param what: message text to log
        """
        if not self.activated:
            return

        # handle indentation of multiple lines
        for i, line in enumerate(filter(None, what.split('\n'))):
            if i == 0:
                message = [self._indent_char * self._indent_width *
                           self.current_level, '<MESSG> ', line.strip()]
            else:
                message = [self._indent_char * self._indent_width *
                           self.current_level, ' '.rjust(8, self._indent_char),
                           line.strip()]
            self._logger.info(''.join(message))

    @contextmanager
    def subproc(self, proc_name: str):
        """
        Log a sub-process using with statement

        :param proc_name: name of the sub-process
        """
        self.begin(proc_name)
        try:
            yield self
            self.ended(proc_name)
        finally:
            pass

    def get_writer(self):
        """
        Get a writer that can redirect sys.stdout to this logger

        :return: writer of this logger
        """

        class LoggerWriter:
            def __init__(self, logger):
                self.logger = logger

            def write(self, message):
                self.logger.message(message)

            def flush(self):
                pass

        return LoggerWriter(self)


def _form_batch_ids(q_dims, batch_size):
    """ form batch ids """
    # grid axes
    batch_id_axes = []  # list of id on each axis
    batch_id_limits = []  # list of full-range slice of each axis
    for q_dim in q_dims:
        batch_id_axis = []
        for q_loc in range(0, q_dim, batch_size):
            q_id = slice(q_loc, min(q_loc + batch_size, q_dim))
            batch_id_axis.append(q_id)
        batch_id_axes.append(batch_id_axis)
        batch_id_limits.append(slice(0, len(batch_id_axis)))

    # flatten by np.mgrid
    id_of_ids = np.mgrid[batch_id_limits].reshape(len(q_dims), -1).T

    # form all batch ids
    batch_ids_all = []
    for id_of_id in id_of_ids:
        batch_id = []
        for iq, q_dim in enumerate(q_dims):
            batch_id.append(batch_id_axes[iq][id_of_id[iq]])
        batch_ids_all.append(tuple(batch_id))
    return batch_ids_all



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# base.py
# ffsas: free-form inversion for small-angle scattering
# Copyright © 2021 SciML, STFC, UK. All rights reserved.


""" base class SASModel """

import os
from abc import abstractmethod

import h5py
import torch

from ffsas import torch_dtype
from ffsas.utils import _form_batch_ids, MultiLevelLogger


class SASModel:
    @classmethod
    @abstractmethod
    def compute_G(cls, q_list, par_dict, const_dict, V=None):
        """
        Compute the Green's tensor `G`

        :param q_list: `list` of scattering vectors `q`
        :param par_dict: `dict` of model parameters as `torch.Tensor`s
        :param const_dict: `dict` of model constants
        :param V: the volume tensor (default=`None`, compute if needed)
        :return: the Green's tensor `G` as a `torch.Tensor`
        """
        pass

    @classmethod
    @abstractmethod
    def get_par_keys_G(cls):
        """
        Get the keys of the parameters used for computing `G`

        :return: `list` of parameters keys ordered by their dimensions in `G`
        """
        pass

    @classmethod
    @abstractmethod
    def compute_V(cls, par_dict):
        """
        Compute the volume `V`

        :param par_dict: `dict` of model parameters as `torch.Tensor`s
        :return: `V` as a `torch.Tensor`
        """
        pass

    @classmethod
    @abstractmethod
    def get_par_keys_V(cls):
        """
        Get the keys of the parameters used for computing `V`

        :return: `list` of parameters keys ordered by their dimensions in `V`
        """
        pass

    @classmethod
    def compute_G_mini_batch(cls, q_list, par_dict, const_dict,
                             fixed_par_weights=None,
                             G_file=None, batch_size=None, device='cpu',
                             log_file=None, log_screen=True):
        """
        Compute the Green's tensor `G` by GPU-accelerated mini-batch computation

        :param q_list: `list` of scattering vectors `q`
        :param par_dict: `dict` of model parameters as `torch.Tensor`s
        :param const_dict: `dict` of model constants
        :param fixed_par_weights: `dict` of fixed parameters and their weights
        :param G_file: write `G` to a HDF5 file named `G_file`, with the
            dateset named `G-TENSOR` (default=None, return `G` as a
            `torch.Tensor`)
        :param batch_size: batch size along each `q`-dimension
            (default=None, compute entire `G` in one batch)
        :param device: computing device (default='cpu')
        :param log_file: save log to this file (default=None)
        :param log_screen: print log to stdout (default=True)
        :return: if `G_file` is `None`, return `G` as a `torch.Tensor`;
            otherwise, return `G_file` as a string
        """
        # logger
        logger = MultiLevelLogger()
        logger.activate(file_path=log_file, screen=log_screen)

        with logger.subproc("Computing the Green's tensor G"):
            logger.message(f'Class name = {cls.__name__}')

            with logger.subproc("Checking tensor dtypes"):
                for q in q_list:
                    assert q.dtype == torch_dtype, \
                        f'dtype of q must be {torch_dtype}'
                for key, par in par_dict.items():
                    assert par.dtype == torch_dtype, \
                        f'dtype of parameter "{key}" must be {torch_dtype}'
                logger.message(f'torch dtype = {torch_dtype}')

            with logger.subproc("Computing volume V"):
                V = cls.compute_V(par_dict)

            with logger.subproc("Handling fixed parameters"):
                if fixed_par_weights is None:
                    fixed_par_weights = {}
                par_keys_free = []
                par_dims_fixed = []
                for loc, key in enumerate(cls.get_par_keys_G()):
                    if key not in fixed_par_weights.keys():
                        par_keys_free.append(key)
                    else:
                        par_dims_fixed.append(loc)
                        assert fixed_par_weights[key].size(0) == par_dict[
                            key].size(0), \
                            f'sizes of weights and coordinates do not match ' \
                            f'for fix parameter "{key}"'
                logger.message(f'model parameters = {cls.get_par_keys_G()}')
                logger.message(
                    f'fixed parameters = {list(fixed_par_weights.keys())}')
                logger.message(f'free parameters = {par_keys_free}')

            with logger.subproc("Creating space for G"):
                # container for G
                q_dims = [q.size(0) for q in q_list]
                G_shape = q_dims + [par_dict[key].size(0) for key in
                                    par_keys_free]
                logger.message(f'G shape = {G_shape}')
                G_count = torch.prod(torch.tensor(G_shape))
                bits = torch.finfo(torch_dtype).bits
                G_size_MB = (G_count * bits) / 8e6
                logger.message(f'G count = {G_count}')
                logger.message(f'G size (MB) = {G_size_MB}')
                if G_file is None:
                    G = torch.empty(G_shape, dtype=q_list[0].dtype)
                    G_h5 = False
                    logger.message('G stored in memory')
                else:
                    os.system(f'rm -f {G_file}')
                    G_h5 = h5py.File(G_file, 'w')
                    G = G_h5.create_dataset('G-TENSOR', G_shape,
                                            dtype=str(q_list[0].dtype)[6:])
                    logger.message(f'G stored in file {G_file}')

            with logger.subproc("Sending parameters and constants to device"):
                # send par_dict to device
                par_dict_device = {}
                for key, par in par_dict.items():
                    par_dict_device[key] = par.to(device)

                # send tensors in const_dict to device
                const_dict_device = {}
                for key, const in const_dict.items():
                    if isinstance(const, torch.Tensor):
                        const_dict_device[key] = const.to(device)
                    else:
                        const_dict_device[key] = const

                # sending fixed weights to device
                fixed_par_weights_device = {}
                for key, weights in fixed_par_weights.items():
                    fixed_par_weights_device[key] = weights.to(device)

                logger.message(f'device = {device}')

            with logger.subproc("Creating batch indices"):
                # create batch ids
                if batch_size is None:
                    batch_size = max(q_dims)
                batch_ids = _form_batch_ids(q_dims, batch_size)
                logger.message(f'batch size = {batch_size}')
                logger.message(f'batch count = {len(batch_ids)}')

            with logger.subproc("Computing G by batches"):
                for i, batch_id in enumerate(batch_ids):
                    q_list_device = []
                    for iq, q in enumerate(q_list):
                        q_list_device.append(q[batch_id[iq]].to(device))
                    batch = cls.compute_G(q_list_device, par_dict_device,
                                          const_dict_device, V=V.to(device))
                    # fixed parameters
                    for dim in par_dims_fixed[::-1]:
                        w = fixed_par_weights_device[cls.get_par_keys_G()[dim]]
                        batch = torch.tensordot(
                            batch, w, dims=[[len(q_dims) + dim], [0]])
                    G[tuple(batch_id)] = batch.to('cpu')
                    logger.message(f'batch {i + 1} / {len(batch_ids)}, '
                                   f'elapsed={logger.elapsed_shallowest:f} sec')

            # return
            if G_h5:
                G_h5.close()
                return G_file
            else:
                return G

    @classmethod
    def compute_average_V(cls, par_dict, w_dict):
        """
        Compute average volume

        :param par_dict: `dict` of model parameters as `torch.Tensor`s
        :param w_dict: `dict` of parameter weights as `torch.Tensor`s
        :return: average volume
        """
        V = cls.compute_V(par_dict)
        for i, key in enumerate(cls.get_par_keys_V()[::-1]):
            V = torch.tensordot(V, w_dict[key], dims=1)
        return V.item()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ellipsoid.py
# ffsas: free-form inversion for small-angle scattering
# Copyright © 2021 SciML, STFC, UK. All rights reserved.


""" ellipsoid model class """

import math

import torch

from ffsas.models.base import SASModel


class Ellipsoid(SASModel):
    @classmethod
    def compute_G(cls, q_list, par_dict, const_dict, V=None):
        # get parameters
        qx, qy = q_list[0], q_list[1]
        rp, re = par_dict['rp'], par_dict['re']
        theta, phi = par_dict['theta'], par_dict['phi']
        drho = const_dict['drho']

        # compute volume
        if V is None:
            V = cls.compute_V(par_dict)

        #############
        # Compute G #
        #############

        # step 1: rotate q
        sin_theta = torch.sin(theta)
        r31 = torch.outer(sin_theta, torch.cos(phi))
        r32 = torch.outer(sin_theta, torch.sin(phi))
        qc = (qx[:, None, None] * r31[None, :, :])[:, None, :, :] + \
             (qy[:, None, None] * r32[None, :, :])[None, :, :, :]
        qa = torch.sqrt(torch.clip(
            (qx ** 2)[:, None, None, None] +
            (qy ** 2)[None, :, None, None] - qc ** 2, min=0.))

        # step 2: qr
        qa_re = torch.moveaxis(qa[:, :, :, :, None] *
                               re[None, None, None, None, :], 4, 2)
        qc_rp = torch.moveaxis(qc[:, :, :, :, None] *
                               rp[None, None, None, None, :], 4, 2)
        qr = torch.sqrt(torch.square(qa_re)[:, :, None, :, :, :] +
                        torch.square(qc_rp)[:, :, :, None, :, :])

        # step 3: shape factor
        shape_factor = 3. * (torch.sin(qr) - qr * torch.cos(qr)) / qr ** 3
        # limit at 0
        shape_factor = torch.nan_to_num(shape_factor,
                                        nan=1., posinf=1., neginf=1.)

        # step 4: G
        G = (drho * shape_factor * V[None, None, :, :, None, None]) ** 2
        return G

    @classmethod
    def get_par_keys_G(cls):
        return ['rp', 're', 'theta', 'phi']

    @classmethod
    def compute_V(cls, par_dict):
        rp, re = par_dict['rp'], par_dict['re']
        return 4. / 3. * math.pi * rp[:, None] * re[None, :] ** 2

    @classmethod
    def get_par_keys_V(cls):
        return ['rp', 're']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# __init__.py
# ffsas: free-form inversion for small-angle scattering
# Copyright © 2021 SciML, STFC, UK. All rights reserved.


""" ffsas module """

import torch

# data type
torch_dtype = torch.float64
torch.set_default_dtype(torch_dtype)


def set_torch_dtype(dtype):
    """ set torch dtype """
    global torch_dtype
    torch_dtype = dtype
    torch.set_default_dtype(torch_dtype)


# specify version here; it will picked up by pip
__version__ = '1.1.1'
